
import carlcm

if __name__ == '__main__':
    c = carlcm.Context()

    c.add_modules(carlcm.DockerModule(),
                  carlcm.ExampleModule(),
                  carlcm.ConsulModule(encrypt='pmsKacTdVOb4x8/Vtr9PWw==',
                                      mode='server', webui=True))

    c.run_modules()




from stat import S_ISDIR, S_IMODE

import fake_filesystem
from nose.tools import *
from mock import Mock, MagicMock, call

import carlcm

c = None

class TestCarlCMContext(object):

    def setup(self):
        global c
        c = carlcm.MockContext()
        self.open = c.open
        self.os = c.os
        with self.open('existingfile', 'wb') as f:
            f.write('asdf')
        self.os.mkdir('existingdir')
        self.os.chmod('existingdir', 0712)
        self.os.chown('existingdir', 37, 75)

    def test_cmd(self):
        eq_(c.cmd(['pwd']), True)
        c._cmd.assert_called_once_with(['pwd'])

    def test_cmd_quiet(self):
        eq_(c.cmd(['pwd'], quiet=True), True)
        c._cmd_quiet.assert_called_once_with(['pwd'])

    def test_shell(self):
        c.cmd = Mock(return_value=True)
        eq_(c.shell('pwd'), True)
        c.cmd.assert_called_once_with('pwd', shell=True)

    def test_mkdir_already_exists(self):
        eq_(c.mkdir('existingdir'), False)

    def test_mkdir_simple_not_exists(self):
        eq_(self.os.path.isdir('/path'), False)
        eq_(c.mkdir('/path'), True)
        eq_(self.os.path.isdir('/path'), True)

    def test_mkdir_complex_not_exists(self):
        eq_(self.os.path.isdir('/pa/th'), False)
        eq_(c.mkdir('/pa/th'), True)
        eq_(self.os.path.isdir('/pa/th'), True)

    def test__apply_permissions_matched_num(self):
        eq_(c._apply_permissions('existingdir', 37, 75, '712'), False)

    def test__apply_permissions_unmatched_uid(self):
        eq_(c._apply_permissions('existingdir', 38, 75, '712'), True)
        eq_(self.os.stat('existingdir').st_uid, 38)

    def test__apply_permissions_unmatched_gid(self):
        eq_(c._apply_permissions('existingdir', 37, 76, '712'), True)
        eq_(self.os.stat('existingdir').st_gid, 76)

    def test__apply_permissions_unmatched_mode(self):
        eq_(c._apply_permissions('existingdir', 37, 76, '715'), True)
        eq_(S_IMODE(self.os.stat('existingdir').st_mode), 0715)

    def test_file_already_exists_and_equal_src_data(self):
        eq_(c.file('existingfile', src_data='asdf'), False)

    def test_file_already_exists_and_equal_src_path(self):
        with self.open('src', 'wb') as f: f.write('asdf')
        eq_(c.file('existingfile', src_path='src'), False)

    def test_file_already_exists_and_unequal_src_data(self):
        eq_(c.file('existingfile', src_data='ff'), True)
        eq_(self.open('existingfile', 'rb').read(), 'ff')

    def test_file_already_exists_and_unequal_src_path(self):
        with self.open('src', 'wb') as f: f.write('ff')
        eq_(c.file('existingfile', src_path='src'), True)
        eq_(self.open('existingfile', 'rb').read(), 'ff')

    def test_file_new_src_data(self):
        eq_(c.file('/file.txt', src_data='asdf'), True)
        eq_(self.open('/file.txt', 'rb').read(), 'asdf')

    def test_file_new_src_path_subdir(self):
        with self.open('file.txt', 'wb') as f: f.write('asdf')
        eq_(c.file('dir/subdir/', src_path='file.txt'), True)
        eq_(self.open('dir/subdir/file.txt', 'rb').read(), 'asdf')

    def test_line_in_file_exists(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', 'blah'), True)
        eq_(self.open('f', 'rb').read(), 'asdf\nblah\n')

    def test_line_in_file_exists_and_matches(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', 'asdf'), False)
        eq_(self.open('f', 'rb').read(), 'asdf\n')

    def test_line_in_file_exists_no_newline(self):
        c.file('f', src_data='asdf')
        eq_(c.line_in_file('f', 'blah'), True)
        eq_(self.open('f', 'rb').read(), 'asdf\nblah\n')

    def test_line_in_file_absent_match(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', 'asdf', state='absent'), True)
        eq_(self.open('f', 'rb').read(), '\n')

    def test_line_in_file_absent_nomatch(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', 'fasdf', state='absent'), False)
        eq_(self.open('f', 'rb').read(), 'asdf\n')

    def test_line_in_file_regexp_match(self):
        c.file('f', src_data='asdf')
        eq_(c.line_in_file('f', regexp='^as', line='blah'), True)
        eq_(self.open('f', 'rb').read(), 'blah\n')

    def test_line_in_file_regexp_match_equal(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', regexp='^as', line='asdf'), False)
        eq_(self.open('f', 'rb').read(), 'asdf\n')

    def test_line_in_file_regexp_nomatch(self):
        eq_(c.line_in_file('existingfile', regexp='^bl', line='blah'), True)
        eq_(self.open('existingfile', 'rb').read(), 'asdf\nblah\n')

    def test_line_in_file_regexp_absent_match(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', regexp='^as', state='absent'), True)
        eq_(self.open('f', 'rb').read(), '\n')

    def test_line_in_file_regexp_absent_nomatch(self):
        c.file('f', src_data='asdf\n')
        eq_(c.line_in_file('f', regexp='^b', state='absent'), False)
        eq_(self.open('f', 'rb').read(), 'asdf\n')

    @raises(ValueError)
    def test_line_in_file_file_nonexistent(self):
        c.line_in_file('nonexistentfile', regexp='^bl', line='blah')

    def test_template_new_src_data(self):
        eq_(c.template('/file.txt', src_data='{{ x }}df', x='as'), True)
        eq_(self.open('/file.txt', 'rb').read(), 'asdf')

    def test_template_new_src_path(self):
        with self.open('src.j2', 'wb') as f: f.write('{{ x }}df')
        eq_(c.template('/file.txt', src_path='src.j2', x='as'), True)
        eq_(self.open('/file.txt', 'rb').read(), 'asdf')

    def test_template_match_src_data(self):
        eq_(c.template('existingfile', src_data='{{ x }}df', x='as'), False)

    def test_triggers(self):
        eq_(c.cmd([], triggered_by='t1'), False)
        eq_(c.cmd([], triggers='t1'), True)
        eq_(c.cmd([], triggered_by='t1'), True)
        eq_(c.cmd([], triggered_by=['t1', 't2']), True)
        eq_(c.cmd([], triggered_by=['t2', 't3']), False)
        eq_(c.cmd([], triggers=['t2','t3']), True)
        eq_(c.cmd([], triggered_by=['t3']), True)
        eq_(c.cmd([], triggered_by='t4', triggers='t5'), False)
        eq_(c.cmd([], triggered_by='t5'), False)
        eq_(c.mkdir('existingdir', triggers='t6'), False)
        eq_(c.cmd([], triggered_by='t6'), False)

    def test__groupadd_cmd(self):
        eq_(c._groupadd_cmd('agroup'), ['groupadd', 'agroup'])
        eq_(c._groupadd_cmd('agroup', gid=45), ['groupadd', '-g', '45', 'agroup'])

    def test__useradd_cmd(self):
        eq_(c._useradd_cmd('auser'), ['useradd', '-U', 'auser'])
        eq_(c._useradd_cmd('auser', gid=45), ['useradd', '-g', '45', '-U', 'auser'])
        eq_(c._useradd_cmd('auser', uid=45), ['useradd', '-u', '45', '-U', 'auser'])
        eq_(c._useradd_cmd('auser', home=False), ['useradd', '-M', '-U', 'auser'])
        eq_(c._useradd_cmd('auser', home='/var/auser'), ['useradd', '-d', '/var/auser', '-U', 'auser'])
        eq_(c._useradd_cmd('auser', shell='/bin/zsh'), ['useradd', '-s', '/bin/zsh', '-U', 'auser'])
        eq_(c._useradd_cmd('auser', comment='blah'), ['useradd', '-c', 'blah', '-U', 'auser'])

    def test_group_exists(self):
        c.groups += [{'name':'group', 'id': 1003}]
        eq_(c.group('group'), False)

    def test_group_new(self):
        eq_(c.group('group'), True)
        eq_(c.groups[1], {'name':'group', 'id': 1000})

    def test_group_new_with_gid(self):
        eq_(c.group('group', gid=74), True)
        eq_(c.groups[1], {'name':'group', 'id': 74})

    def test_user_exists(self):
        c.users += [{'name':'jessie', 'id':1002, 'home':None}]
        eq_(c.user('jessie', home=False), False)

    def test_user_new(self):
        eq_(c.user('jessie'), True)
        eq_(c.users[1]['name'], 'jessie')
        eq_(c.users[1]['groups'], ['jessie'])
        eq_(c.groups[1]['name'], 'jessie')
        eq_(self.os.path.isdir('/home/jessie'), True)
        eq_(S_IMODE(self.os.stat('/home/jessie').st_mode), 0755)
        eq_(self.os.stat('/home/jessie').st_uid, 1000)
        eq_(self.os.stat('/home/jessie').st_gid, 1000)

    def test_user_new_homeless(self):
        eq_(c.user('jessie', home=False), True)
        eq_(self.os.path.isdir('/home/jessie'), False)

    def test_user_new_homed(self):
        eq_(c.user('jessie', home='/var/jessie', uid=31, gid=76, home_mode='750'), True)
        eq_(self.os.path.isdir('/home/jessie'), False)
        eq_(self.os.path.isdir('/var/jessie'), True)
        eq_(S_IMODE(self.os.stat('/var/jessie').st_mode), 0750)
        eq_(self.os.stat('/var/jessie').st_uid, 31)
        eq_(self.os.stat('/var/jessie').st_gid, 76)

    def test_user_new_grouped(self):
        eq_(c.user('jessie', groups=['agroup', 'bgroup']), True)
        eq_(c.users[1]['name'], 'jessie')
        eq_(c.users[1]['groups'], ['agroup', 'bgroup', 'jessie'])

    def test_user_authorized_keys(self):
        eq_(c.user('jessie', authorized_keys='ssh-rsa blah== blah@blah'), True)
        eq_(c._read_file('/home/jessie/.ssh/authorized_keys'),
            'ssh-rsa blah== blah@blah')
        eq_(S_IMODE(self.os.stat('/home/jessie/.ssh').st_mode), 0700)
        eq_(S_IMODE(self.os.stat('/home/jessie/.ssh/authorized_keys').st_mode), 0600)
        eq_(c.user('jessie', authorized_keys=['ssh-rsa blah== blah@blah']), False)
        eq_(c.user('jessie', authorized_keys=None), False)
        eq_(c.user('jessie', authorized_keys='x'), True)
        eq_(c._read_file('/home/jessie/.ssh/authorized_keys'), 'x')
        eq_(c.user('jessie', authorized_keys=['a','b','c']), True)
        eq_(c._read_file('/home/jessie/.ssh/authorized_keys'), 'a\nb\nc')

    @raises(AssertionError)
    def test_user_authorized_keys_homeless(self):
        c.user('jessie', authorized_keys='ssh-rsa blah== blah@blah', home=False)

    def test_user_groups_unchanged(self):
        c.users += [{'name':'jessie', 'id':1000,
                     'groups':['wheel', 'admins', 'gamers', 'jessie']}]
        eq_(c.user('jessie', home=False, groups=['wheel', 'admins', 'gamers']), False)

    def test_user_groups_changed(self):
        c.users += [{'name':'jessie', 'id':1000,
                     'groups':['devs', 'admins', 'gamers', 'jessie']}]
        eq_(c.user('jessie', home=False, groups=['wheel', 'admins', 'gamers']), True)
        eq_(c.users[1]['groups'], ['admins', 'gamers', 'jessie', 'wheel'])

    def test_current_packages(self):
        c._cmd_quiet.side_effect = ['''
Desired=Unknown/Install/Remove/Purge/Hold
| Status=Not/Inst/Conf-files/Unpacked/halF-conf/Half-inst/trig-aWait/Trig-pend
|/ Err?=(none)/Reinst-required (Status,Err: uppercase=bad)
||/ Name                                                  Version                                             Architecture Description
+++-=====================================================-===================================================-============-===============================================================================
ii  ack-grep                                              2.12-1                                              all          grep-like program specifically for large source trees
ii  acpid                                                 1:2.0.21-1ubuntu2                                   amd64        Advanced Configuration and Power Interface event daemon
''']
        eq_(c.current_packages(), {'ack-grep':'2.12-1','acpid':'1:2.0.21-1ubuntu2'})
        eq_(c.current_packages(), {'ack-grep':'2.12-1','acpid':'1:2.0.21-1ubuntu2'})
        c._cmd_quiet.assert_has_calls([call(['dpkg', '-l'])])

    def test_packages_new(self):
        c.current_packages = Mock(return_value={})
        eq_(c.packages(['ack-grep', 'acpid']), True)
        c._cmd_quiet.assert_called_once_with(['apt-get', 'install', '-y', 'ack-grep', 'acpid'])

    def test_packages_present(self):
        c.current_packages = Mock(return_value={'ack-grep':'2.12-1','acpid':'1:2.0.21-1ubuntu2'})
        eq_(c.packages(['ack-grep', 'acpid']), False)
        c._cmd_quiet.assert_has_calls([])

    def test_packages_mixed(self):
        c.current_packages = Mock(return_value={'ack-grep':'2.12-1'})
        eq_(c.packages(['ack-grep', 'acpid']), True)
        c._cmd_quiet.assert_called_once_with(['apt-get', 'install', '-y', 'acpid'])

    def test_download_new(self):
        c.mock_urls['http://blah.com/blah.txt'] = 'asdf\n'
        eq_(c.download('/blah.txt', 'http://blah.com/blah.txt'), True)
        eq_(c._read_file('/blah.txt'), 'asdf\n')

    def test_download_new_hash_match(self):
        c.mock_urls['http://blah.com/blah.txt'] = 'asdf\n'
        eq_(c.download('/blah.txt', 'http://blah.com/blah.txt',
                       md5='2b00042f7481c7b056c4b410d28f33cf'), True)
        eq_(c._read_file('/blah.txt'), 'asdf\n')

    @raises(AssertionError)
    def test_download_new_hash_mismatch(self):
        c.mock_urls['http://blah.com/blah.txt'] = 'asdf\n'
        c.download('/blah.txt', 'http://blah.com/blah.txt',
                   md5='2b10042f7481c7b056c4b410d28f33cf')

    def test_download_exists(self):
        # NOTE: doesn't even download!
        c._write_file('/blah.txt', 'asdf\n')
        eq_(c.download('/blah.txt', 'http://blah.com/blah.txt'), False)
        eq_(c._read_file('/blah.txt'), 'asdf\n')

    def test_download_exists_hash_match(self):
        c._write_file('/blah.txt', 'asdf\n')
        eq_(c.download('/blah.txt', 'http://blah.com/blah.txt',
                       md5='2b00042f7481c7b056c4b410d28f33cf'), False)

    def test_download_exists_hash_mismatch(self):
        c.mock_urls['http://blah.com/blah.txt'] = 'asdf\n'
        c._write_file('/blah.txt', 'asdff\n')
        eq_(c.download('/blah.txt', 'http://blah.com/blah.txt',
                       md5='2b00042f7481c7b056c4b410d28f33cf'), True)
        eq_(c._read_file('/blah.txt'), 'asdf\n')




import errno
import filecmp
import grp
import hashlib
import json
import os as real_os
import pwd
import re
import shutil
from stat import S_IMODE, S_ISDIR
import subprocess
import sys
import types
import urllib

class Context(object):
    '''
    Most methods return True if something was modified, and False otherwise
    '''

    is_mock = False

    def __init__(self, _os=None, _open=None):
        self.os = _os or real_os
        self.open = _open or open
        self.triggers = set()
        self.package_cache = None
        self.modules = []
        self.actions = {}
        self.action_modules = {}
        self.add_action_module('carlcm.actions.core')

    def _before(self, triggered_by):
        '''
        If this returns True, we should skip the run
        '''
        if triggered_by is None:
            return False
        if type(triggered_by) == str:
            triggered_by = [triggered_by]
        return len(set(triggered_by).intersection(self.triggers)) == 0

    def _after(self, is_new, triggers):
        bad_triggers_message = Exception('triggers must be None, a string, or a list of strings')
        if triggers is None:
            return is_new
        if type(triggers) == str:
            triggers = [triggers]
        if hasattr(triggers, '__iter__'):
            for trigger in triggers:
                if type(trigger) != str:
                    raise bad_triggers_message
        else:
            raise bad_triggers_message
        if is_new:
            self.triggers = self.triggers.union(set(triggers))
        return is_new

    def add_action_module(self, am_name, *args, **kwargs):
        import_name = am_name
        if self.is_mock:
            import_name += '_mock'
        if am_name in self.action_modules:
            return
        from carlcm.actions.action_module import ActionModule
        __import__(import_name)
        py_module = sys.modules[import_name]
        name = py_module.__name__.split('.')[-1]
        classname = ''.join([a.capitalize() for a in name.split('_')])
        am = py_module.__getattribute__(classname)(context=self, *args, **kwargs)
        self.action_modules[name] = am
        for k in dir(am):
            v = am.__getattribute__(k)
            if k[:1] != '_' and type(v) == types.MethodType:
                self.actions[k] = v

    def __getattr__(self, name):
        if name in self.action_modules:
            return self.action_modules[name]
        if name in self.actions:
            action = self.actions[name]
            def _missing(*args, **kwargs):
                triggers = kwargs.get('triggers')
                triggered_by = kwargs.get('triggered_by')
                if 'triggers' in kwargs: del kwargs['triggers']
                if 'triggered_by' in kwargs: del kwargs['triggered_by']
                if self._before(triggered_by):
                    return False
                changed = action(*args, **kwargs)
                return self._after(changed, triggers)
            return _missing
        raise AttributeError('No attribute %s' % name)

    def shell(self, cmd, **kwargs):
        return self.cmd(cmd, shell=True, **kwargs)

    def _cmd_quiet(self, *args, **kwargs):
        return subprocess.check_output(*args, **kwargs)

    def _cmd(self, *args, **kwargs):
        return subprocess.check_call(*args, **kwargs)

    def _cmd_in(self, cmd, stdin, **kwargs):
        proc=subprocess.Popen(cmd,
                              stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT,
                              **kwargs)
        proc.stdin.write(str(stdin))
        proc.stdin.flush()
        stdout, stderr = proc.communicate()
        return stdout

    def current_packages(self):
        if self.package_cache is not None:
            return self.package_cache
        s = self._cmd_quiet(['dpkg', '-l'])
        d = {}
        for line in [l.strip().split() for l in s.split("\n") if l.strip()[:2] == 'ii']:
            if len(line) >= 3:
                d[line[1]] = line[2]
        self.package_cache = d
        return d

    def package_manager_update(self, triggers=None, triggered_by=None):
        if self._before(triggered_by): return False
        self._cmd_quiet(['apt-get', 'update'])
        return self._after(True, triggers)

    def package(self, package, **kwargs):
        return self.packages([package], **kwargs)

    def packages(self, packages, triggers=None, triggered_by=None):
        if self._before(triggered_by): return False
        new_packages = [p for p in packages if p not in self.current_packages()]
        if len(new_packages) > 0:
            old_env = self.os.getenv('DEBIAN_FRONTEND', None)
            self.os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
            self._cmd_quiet(['apt-get', 'install', '-y'] + sorted(new_packages))
            if old_env:
                self.os.environ['DEBIAN_FRONTEND'] = old_env
            else:
                del self.os.environ['DEBIAN_FRONTEND']
            self.package_cache = None
        return self._after(len(new_packages) > 0, triggers)

    def cmd(self, cmd, quiet=False, triggers=None, triggered_by=None, **kwargs):
        if self._before(triggered_by): return False
        if quiet:
            self._cmd_quiet(cmd, **kwargs)
        else:
            self._cmd(cmd, **kwargs)
        return self._after(True, triggers)

    def _mkdir(self, d):
        """
        Based on http://code.activestate.com/recipes/82465-a-friendly-mkdir/
        """
        d = self.os.path.realpath(d)
        id = self.os.path.isdir(d)
        if id:
            return False
        elif self.os.path.isfile(d):
            raise OSError("file exists: " % d)
        else:
            h, t = self.os.path.split(d)
            if h: self._mkdir(h)
            if t: self.os.mkdir(d)
        return True

    def _user_name_to_uid(self, user):
        try:
            return pwd.getpwnam(user).pw_uid
        except KeyError:
            return None
    def _group_name_to_gid(self, group):
        try:
            return grp.getgrnam(group).gr_gid
        except KeyError:
            return None
    def _user_home(self, user):
        try:
            return pwd.getpwnam(user).pw_dir
        except KeyError:
            return None

    def _apply_permissions(self, path, owner, group, mode):
        stat = self.os.stat(path)
        matched = True
        if mode is not None:
            if type(mode) == str:
                mode = int(mode, 8)
            self.os.chmod(path, mode)
            if S_IMODE(stat.st_mode) != mode:
                matched = False
        if owner is not None or group is not None:
            owner = owner or -1
            group = group or -1
            if type(owner) == str:
                owner = self._user_name_to_uid(owner)
                if owner is None:
                    raise ValueError('no such user!')
            if type(group) == str:
                group = self._group_name_to_gid(group)
                if group is None:
                    raise ValueError('no such group!')
            self.os.chown(path, owner, group)
            if owner >= 0 and stat.st_uid != owner or group >= 0 and stat.st_gid != group:
                matched = False
        return not matched

    def _touch(self, path):
        self._mkdir(self.os.path.dirname(path))
        with self.open(path, 'a') as f:
            f.write('')

    def _read_file(self, path):
        return self.open(path, 'rb').read()

    def _write_file(self, path, data):
        with self.open(path, 'wb') as f:
            f.truncate()
            f.write(data)

    def _urlretrieve(self, url, path):
        urllib.urlretrieve(url, path)

    def _hash_file(self, path, hash_algo):
        m = hashlib.new(hash_algo)
        with self.open(path, 'rb') as f:
            while True:
                s = f.read(4096)
                m.update(s)
                if len(s) <= 0:
                    break
        return m.hexdigest()

    def download(self, path, url,
                 owner=None, group=None, mode=None,
                 triggers=None, triggered_by=None, **kwargs):
        # NOTE: this downloads in-place currently, which is not great
        # for sanctity of the file.

        # Also... right now, if you don't pass a hash, it will only
        # download once, and never check again.  Is that desired?
        # Should it always download and compare the file?
        if self._before(triggered_by): return False
        file_new = not self.os.path.isfile(path)
        if file_new:
            self._touch(path)
        perm_change = self._apply_permissions(path, owner, group, mode)
        if not file_new:
            for a, k in [(a, a+s) for a in hashlib.algorithms for s in ['','sum']]:
                if k in kwargs:
                    file_new = file_new or self._hash_file(path, a) != kwargs[k]

        if file_new:
            self._urlretrieve(url, path)
            for a, k in [(a, a+s) for a in hashlib.algorithms for s in ['','sum']]:
                if k in kwargs:
                    assert self._hash_file(path, a) == kwargs[k]
        return self._after(perm_change or file_new, triggers)

    def file(self, dest_path, src_path=None, src_data=None,
             owner=None, group=None, mode=None,
             triggers=None, triggered_by=None):
        # NOTE: this writes the file in-place currently, which is not
        # great for the stability of reads to that file.
        assert bool(src_path is not None) != bool(src_data is not None)
        if self._before(triggered_by): return False
        if dest_path[-1:] == '/' and src_path:
            _, tail = self.os.path.split(src_path)
            dest_path += tail
        dest_path = self.os.path.realpath(dest_path)
        file_existed = self.os.path.isfile(dest_path)
        if not file_existed:
            self._touch(dest_path)
        perm_change = self._apply_permissions(dest_path, owner, group, mode)
        old_contents = self._read_file(dest_path)
        if type(src_data) is dict:
            src_data = json.dumps(src_data, sort_keys=True,
                                  indent=4, separators=(',', ': ')) + '\n'
        if src_path is not None:
            src_data = self._read_file(src_path)
        contents_match = src_data == old_contents
        if not contents_match:
            self._write_file(dest_path, src_data)
        return self._after(perm_change or not (file_existed and contents_match), triggers)

    def template(self, dest_path, src_path=None, src_data=None,
                 template_parameters=None, engine='jinja2',
                 triggers=None, triggered_by=None, **kwargs):
        if engine == 'jinja2':
            return self.jinja2(dest_path, src_path, src_data, template_parameters,
                               triggers, triggered_by, **kwargs)
        raise ValueError('no matching template engine!')

    def jinja2(self, dest_path, src_path=None, src_data=None,
               owner=None, group=None, mode=None,
               template_parameters=None, triggers=None, triggered_by=None, **kwargs):
        import jinja2
        if self._before(triggered_by): return False
        template_parameters = template_parameters or {}
        for k, v in kwargs.items():
            template_parameters[k] = template_parameters.get(k, v)
        if src_path is not None and src_data is None:
            src_data = self._read_file(src_path)
        # todo: also pass perms
        return self.file(dest_path = dest_path,
                         owner=owner, group=group, mode=mode,
                         src_data = jinja2.Template(src_data).render(template_parameters),
                         triggers = triggers)

    def _groupadd_cmd(self, groupname, gid=None):
        cmd = ['groupadd']
        if gid is not None:
            cmd += ['-g', str(gid)]
        cmd += [groupname]
        return cmd

    def _groupadd(self, groupname, gid=None):
        self._cmd_quiet(self._groupadd_cmd(groupname, gid))

    def group(self, groupname, gid=None, triggers=None, triggered_by=None):
        if self._before(triggered_by): return False
        group_existed = bool(self._group_name_to_gid(groupname))
        if not group_existed:
            self._groupadd(groupname, gid)
        return self._after(not group_existed, triggers)

    def _useradd_cmd(self, username, home=True, uid=None, gid=None,
                     groups=None, shell=None, comment=None):
        cmd = ['useradd']
        if home is not True:
            if home is False:
                cmd += ['-M']
            elif type(home) is str:
                cmd += ['-d', home]
            else:
                raise ValueError('errrrrr')
        if uid is not None:
            cmd += ['-u', str(uid)]
        if gid is not None:
            cmd += ['-g', str(gid)]
        if shell is not None:
            cmd += ['-s', str(shell)]
        if comment is not None:
            cmd += ['-c', str(comment)]
        cmd += ['-U', username]
        return cmd

    def _useradd(self, username, home=True, uid=None, gid=None,
                 groups=None, shell=None, comment=None):
        self._cmd_quiet(self._useradd_cmd(username, home, uid, gid,
                                          groups, shell, comment))

    def _user_groups(self, username):
        gs = self._cmd_quiet(['groups', username])
        return sorted(gs.split(':')[-1].strip().split())

    def _add_user_to_group(self, username, groupname):
        self._cmd_quiet(['gpasswd', '-a', username, groupname])
    def _remove_user_from_group(self, username, groupname):
        self._cmd_quiet(['gpasswd', '-d', username, groupname])

    def authorized_keys(self, user, authorized_keys, triggers=None, triggered_by=None):
        if self._before(triggered_by): return False
        if type(authorized_keys) is list:
            authorized_keys = '\n'.join(authorized_keys)
        home = self._user_home(user)
        if home is None:
            raise AssertionError(user + ' has no home, but authorized_keys was specified!')
        ssh_dir = self.os.path.join(home, '.ssh')
        self.mkdir(ssh_dir, owner=user, group=user, mode=0700)
        filename = self.os.path.join(ssh_dir, 'authorized_keys')
        changed = self.file(filename, src_data = authorized_keys,
                            owner=user, group=user, mode=0600)
        return self._after(changed, triggers)

    def user(self, username, password=None, encrypted_password=None,
             authorized_keys=None,
             home=True, home_mode='755', uid=None, gid=None, groups=None, shell=None,
             comment=None, random_password=False, triggers=None, triggered_by=None):
        '''
        http://serverfault.com/questions/367559/
        echo "P4sSw0rD" | openssl passwd -1 -stdin
        '''
        if self._before(triggered_by): return False
        user_existed = bool(self._user_name_to_uid(username))
        if not user_existed:
            self._useradd(username, home, uid, gid,
                          groups, shell, comment)
        home_changed = False
        home_perm_changed = False
        if home is not False:
            _home = self._user_home(username)
            home_changed = self._mkdir(_home)
            home_perm_changed = self._apply_permissions(_home, username, username, home_mode)

        changing_groups = False
        if groups:
            gs = self._user_groups(username)
            existing_groups = set(self._user_groups(username)) - set([username])
            changing_groups = existing_groups != set(groups)
            for group in sorted(list(set(groups) - set(existing_groups))):
                self._add_user_to_group(username, group)
            for group in sorted(list(set(existing_groups) - set(groups))):
                self._remove_user_from_group(username, group)

        # TODO: need to be able to check if the password was the same or not
        if random_password:
            password = real_os.urandom(20).encode('hex')
        if password is not None:
            self._cmd_in(['chpasswd'], username + ':' + password + '\n')
        if encrypted_password is not None:
            self._cmd_in(['chpasswd', '-e'], username + ':' + encrypted_password + '\n')

        changed_auth_keys = False
        if authorized_keys is not None:
            changed_auth_keys = self.authorized_keys(username, authorized_keys)
        return self._after(not user_existed or home_changed or home_perm_changed or changing_groups or changed_auth_keys, triggers)

    def line_in_file(self, path, line=None, regexp=None, state='present',
                     enforce_trailing_newline=True,
                     triggers=None, triggered_by=None):
        if self._before(triggered_by): return False
        if not self.os.path.isfile(path):
            raise ValueError('path %s is not a file!' % path)
        assert state in ['present', 'absent']
        if state == 'present': assert line is not None
        if state == 'absent': assert bool(regexp is not None) != bool(line is not None)
        if type(regexp) is str:
            regexp = re.compile(regexp)
        data = self._read_file(path)
        lines = data.split('\n')
        line_ix = None
        if regexp is not None:
            for ix in xrange(len(lines)):
                if regexp.search(lines[ix]):
                    line_ix = ix
        if line_ix is None:
            for ix in xrange(len(lines)):
                if lines[ix] == line:
                    line_ix = ix
        if state == 'present':
            assert line is not None
            if line_ix is None:
                line_ix = len(lines)
                if len(lines) > 0 and lines[-1] == '':
                    line_ix -= 1
            assert line_ix <= len(lines)
            if line_ix == len(lines):
                lines += ['']
            lines[line_ix] = line
        if state == 'absent' and line_ix is not None:
            lines = lines[:line_ix] + lines[line_ix+1:]
        data2 = '\n'.join(lines)
        if enforce_trailing_newline and (len(data2) == 0 or data2[-1] != '\n'):
            data2 += '\n'
        return self.file(path, src_data=data2)

    def add_modules(self, *args):
        self.modules += args
        return self

    def run_modules(self):
        packages = []
        for module in self.modules:
            packages += module.packages()
        # TODO: change the package format to allow inclusion of versions... somehow
        self.packages(packages)
        for module in self.modules:
            module.main(self)
        return self

# TODO: rsync, git repo, apt sources, apt keys, ssh authorized_keys, cron

class MockContext(Context):

    is_mock = True

    def __init__(self, fs=None, users=None, groups=None):
        import fake_filesystem
        from mock import Mock

        self.fs = fs or fake_filesystem.FakeFilesystem()
        self.users = users or [{'name':'root', 'id':0,
                                'groups':['root'], 'home':'/root'}]
        self.groups = groups or [{'name':'root', 'id':0}]
        self.mock_urls = {}
        self._cmd = Mock()
        self._cmd_quiet = Mock()
        Context.__init__(self,
                         _os=fake_filesystem.FakeOsModule(self.fs),
                         _open=fake_filesystem.FakeFileOpen(self.fs))
    def _cmd(self):
        assert False
    def _cmd_quiet(self):
        assert False
    def _cmd_in(self):
        assert False
    def _user_home(self, user):
        return [x for x in self.users if x['name'] == user][0].get('home')
    def _user_groups(self, user):
        return [x for x in self.users if x['name'] == user][0].get('groups', [])
    def _user_name_to_uid(self, user):
        try: return [x for x in self.users if x['name'] == user][0]['id']
        except IndexError: return None
    def _group_name_to_gid(self, group):
        try: return [x for x in self.groups if x['name'] == group][0]['id']
        except IndexError: return None
    def _next_id(self, arr):
        i = 1000
        for item in arr:
            i = max(item['id'] + 1, i)
        return i

    def _groupadd(self, groupname, gid=None):
        gid = gid or self._next_id(self.groups)
        if len([x for x in self.groups if x['id'] == gid]) > 0:
            raise Exception('gid already taken!')
        self.groups += [{'name':groupname, 'id': gid}]

    def _useradd(self, username, home=True, uid=None, gid=None,
                 groups=None, shell=None, comment=None):
        uid = uid or self._next_id(self.groups)
        if len([x for x in self.users if x['id'] == uid]) > 0:
            raise Exception('uid already taken!')
        user = {'name':username, 'id':uid, 'comment':comment, 'shell':shell,
                'home':'/home/'+username, 'groups':[username]}
        if home is not True:
            if home is False:
                user['home'] = None
            elif type(home) is str:
                user['home'] = home
            else:
                raise ValueError('errrrrr')
        self.users += [user]
        self._groupadd(username, gid)

    def _add_user_to_group(self, username, groupname):
        user = [x for x in self.users if x['name'] == username][0]
        user['groups'] = sorted(list(set(user['groups'] + [groupname])))

    def _remove_user_from_group(self, username, groupname):
        user = [x for x in self.users if x['name'] == username][0]
        user['groups'] = sorted(list(set(user['groups']) - set([groupname])))

    def _urlretrieve(self, url, path):
        self._write_file(path, self.mock_urls[url])




from .instance_role import InstanceRole

class Example(InstanceRole):

    def main(self, context):
        super(Example, self).main(context)
        context.package('nginx')
        context.mkdir('/usr/share/nginx/html')
        context.file('/etc/nginx/sites-available/default', src_data='''
server {
        listen   80 default;
        server_name  localhost;

        location / {
                root   /usr/share/nginx/html;
                index  index.html index.htm;
        }
}
''', triggers='nginx')
        context.file('/usr/share/nginx/html/index.html', src_data='''
<html>
<body>
Hello World!
</body>
</html>
''')
        context.cmd(['service', 'nginx', 'reload'], triggered_by='nginx')
        return




import json
import time

import boto
import boto.ec2
import boto.iam
import boto.utils
import consul

from .action_module import ActionModule

class Aws(ActionModule):

    def __init__(self, context=None, region=None, ec2=None, iam=None, in_ec2=False):
        self.context = context
        self.in_ec2 = in_ec2
        self.ec2 = ec2
        self.iam = iam
        self.meta = {}
        self.region = region
        self._connect()

    def _connect(self):
        if self.in_ec2:
            self.meta = boto.utils.get_instance_metadata()
        if self.region is None:
            if self.in_ec2:
                self.region = self.meta['hostname'].split('.')[1]
            else:
                self.region = 'us-west-2'

        self.ec2 = self.ec2 or boto.ec2.connect_to_region(self.region)
        self.iam = self.iam or boto.connect_iam()
        assert self.ec2
        assert self.iam

    def iam_role_and_profile(self, name):
        self.iam_role(name)
        self.instance_profile(name, name)

    def iam_role(self, name, policy_name='managedpolicy', policy_data={}):
        if type(policy_data) != str:
            policy_data = json.dumps(policy_data, sort_keys=True, indent=4, separators=(',', ': '))
        try: role = self.iam.get_role(name)
        except boto.exception.BotoServerError: role = None
        if role is None:
            self.iam.create_role(name)
            time.sleep(1)
            role = self.iam.get_role(name)
        # ...someday, maybe support for multiple policies via iam.list_role_policies()
        self.iam.put_role_policy(name, policy_name, policy_data)

    def instance_profile(self, name, role_name=None):
        '''
        If role_name is None, ensure the instance profile has no role
        associated with it, otherwise, the appropriate role.
        '''
        try: profile = self.iam.get_instance_profile(name)
        except boto.exception.BotoServerError: profile = None
        if profile is None:
            self.iam.create_instance_profile(name)
            time.sleep(1)
            profile = self.iam.get_instance_profile(name)
        roles = profile['get_instance_profile_response']['get_instance_profile_result']['instance_profile']['roles']
        should_add = False
        should_remove = False
        if role_name is None:
            if len(roles) != 0:
                should_remove = True
        if role_name is not None:
            if len(roles) == 0:
                should_add = True
            elif roles['member']['role_name'] != role_name:
                should_remove = True
                should_add = True
        if should_remove:
            self.iam.remove_role_from_instance_profile(name, roles['member']['role_name'])
        if should_add:
            self.iam.add_role_to_instance_profile(name, role_name)

        if role_name is not None and len(roles) != 0:
            should_add = True

    def security_group(self, name, description):
        try: group = self.ec2.get_all_security_groups(groupnames=[name])[0]
        except boto.exception.EC2ResponseError: group = None
        if group is None:
            self.ec2.create_security_group(name, description)
            group = self.ec2.get_all_security_groups(groupnames=[name])[0]
            return True
        # TODO: permissions on group? Or maybe have that separately?
        return False

    def _get_security_group(self, name):
        return self.ec2.get_all_security_groups(groupnames=[name])[0]

    def _security_group_id(self, name):
        return self._get_security_group(name).id

    def launch_instance(self, name, security_groups, iam_role=None, **kwargs):
        insts = self._get_instances()
        if len([i for i in insts if i.tags.get('Name') == name and i.state != 'terminated']) > 0:
            raise Exception('An instance with the name "%s" already exists!' % (name))
        if type(security_groups) == str:
            security_groups = [security_groups]
        security_group_ids = []
        for sg in security_groups:
            if len(sg) == 11 and sg[:3] == 'sg-':
                security_group_ids.append(sg)
            else:
                security_group_ids.append(self._security_group_id(sg))
        kwargs['security_group_ids'] = security_group_ids
        kwargs['min_count'] = kwargs.get('min_count', 1)
        kwargs['max_count'] = kwargs.get('max_count', 1)
        if iam_role is not None:
            kwargs['instance_profile_name'] = iam_role
        kwargs['image_id'] = kwargs.get('image_id', 'ami-870a2fb7')

        reservation = self.ec2.run_instances(**kwargs)
        # TODO: loop 'til the instances are available in the api, due to eventual consistency?
        self.ec2.create_tags([inst.id for inst in reservation.instances],
                        {'Name': name})
        return True

    def _get_instances(self):
        return reduce(lambda a, b: a + b, [a.instances for a in self.ec2.get_all_reservations()], [])




from ..internal_utils import *

from .meta_role import MetaRole

class InstanceRole(MetaRole):

    def name_base(self):
        return self.environment_name + '-' + camel_to_hyphen(self.__class__.__name__)

    def security_groups(self):
        return [self.name_base()]

    def iam_role(self):
        return self.name_base()

    def name(self, num):
        return self.name_base() + '-%04i' % num

    def user_data(self):
        return '''#!/bin/bash
sudo apt-get update
sudo apt-get install -y python-setuptools
sudo easy_install pip
sudo pip install carlcm
sudo carlcm-run-role '%s' '%s'
''' % (self.import_name(), self.environment_name)

    def instances(self, context):
        insts = context.aws._get_instances()
        sg = self.name_base()
        return [i for i in insts if sg in [g.name for g in i.groups]]

    def next_name(self, context, instances=None):
        instances = instances or self.instances(context)
        names = set([i.tags.get('Name') for i in instances])
        for i in xrange(1, 1000):
            if self.name(i) in names:
                continue
            return self.name(i)
        raise Exception('more than 1000 instances!')

    def launch_one(self, context):
        d = {
            'image_id': 'ami-870a2fb7',
            'key_name': 'carl-ssh-2015-02-14',
            'user_data': self.user_data(),
            'subnet_id': 'subnet-7af8590d',
            'instance_type': 't2.micro',
            'min_count': 1,
            'max_count': 1,
        }
        name = self.next_name(context)
        context.launch_instance(name, security_groups=self.security_groups(), iam_role=self.iam_role(), **d)

    def cluster(self, context):

        for sg in self.security_groups():
            context.security_group(sg, sg)

        if self.iam_role():
            context.iam_role_and_profile(self.iam_role())

        instances = self.instances(context)
        if len(instances) < self.min_count:
            for i in xrange(self.min_count - len(instances)):
                self.launch_one(context)

        return




from ..internal_utils import *

class MetaRole(object):
    name = 'metarole'

    min_count = 1
    max_count = 1

    def import_name(self):
        return 'carlcm.roles.' + camel_to_snake(self.__class__.__name__)

    def __init__(self, environment_name):
        self.environment_name = environment_name

    def gather_information(self):
        '''
        Gathers information about the role from the Consul servers.
        Used before running cluster().
        '''
        return

    def cluster(self, context):
        '''
        Set up and/or maintain the cluster of instances/containers/whatever.
        '''
        return

    def main(self, context):
        '''
        Code that runs on the machine to set up the role.
        '''
        return




import re

def camel_to_snake(name):
    # based on http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def camel_to_hyphen(name):
    return camel_to_snake(name).replace('_','-')




import jinja2

from .action_module import ActionModule

class Core(ActionModule):

    def mkdir(self, path, owner=None, group=None, mode=None):
        path = self.context.os.path.realpath(path)
        is_new = self.context._mkdir(path)
        perm_change = self.context._apply_permissions(path, owner, group, mode)
        return is_new or perm_change



'''

Installs Consul.

If webui is True, it will enable the web ui, visible at 'http://localhost:8500/ui'

'''

from .base import BaseModule

# TODO: tls
# https://www.digitalocean.com/community/tutorials/how-to-secure-consul-with-tls-encryption-on-ubuntu-14-04

# TODO: configurable consul version number?

# TODO: figure out how to put service definitions into this
# http://www.consul.io/docs/agent/services.html
# also, changing service definitions should only send a SIGHUP, not restart the process
# ...maybe service definitions should be available at the Module level, so there's no race between consul and other modules' instantiations

class ConsulModule(BaseModule):

    def __init__(self, encrypt=None, mode='client', servers=None, webui=False,
                 datacenter=None, bootstrap_expect=1):
        assert mode in ['client', 'server']
        self.mode = mode
        self.servers = servers
        self.webui = webui
        self.datacenter = datacenter or 'dc1'
        self.encrypt = encrypt
        self.bootstrap_expect = bootstrap_expect

    def packages(self):
        return ['unzip']

    def main(self, context):
        context.user('consul', home='/var/consul', home_mode='750')
        self._acquire_consul(context)
        if self.webui:
            self._acquire_webui(context)

        if self.mode == 'server':
            config_json = self._server_config()
        else:
            config_json = self._client_config()

        context.file('/etc/consul.d/config.json',
                     owner='root', group='consul', mode='640',
                     src_data=config_json, triggers='consul')
        context.file('/etc/init/consul.conf',
                     src_data=self._upstart_conf(), triggers='consul-restart')

        context.shell('service consul restart', triggered_by='consul-restart')
        context.shell('service consul reload', triggered_by='consul')

    def _acquire_consul(self, context):
        is_new = context.download('/opt/consul/0.4.1/consul.zip',
                                  'https://dl.bintray.com/mitchellh/consul/0.4.1_linux_amd64.zip',
                                  sha256sum='2cf6e59edf348c3094c721eb77436e8c789afa2c35e6e3123a804edfeb1744ac')
        if is_new:
            context.cmd(['unzip', '/opt/consul/0.4.1/consul.zip', '-d', '/opt/consul/0.4.1/'], quiet=True)

    def _acquire_webui(self, context):
        context.mkdir('/opt/consul/0.4.1/web', owner='consul', group='consul')
        ui_is_new = context.download('/opt/consul/0.4.1/web/web.zip',
                                     'https://dl.bintray.com/mitchellh/consul/0.4.1_web_ui.zip',
                                     sha256sum='e02929ed44f5392cadd5513bdc60b7ab7363d1670d59e64d2422123229962fa0')
        if ui_is_new:
            context.cmd(['unzip', '/opt/consul/0.4.1/web/web.zip', '-d', '/opt/consul/0.4.1/web/'], quiet=True)

    def _common_config(self):
        d = {'datacenter':self.datacenter,
             'data_dir':'/var/consul',
             'log_level':'INFO', 'enable_syslog':True}
        if self.encrypt is not None:
            d['encrypt'] = encrypt
        if self.webui:
            d["ui_dir"] = "/opt/consul/0.4.1/web/dist"
            d["client_addr"] = "127.0.0.1"
            d["addresses"] = {
                "dns": "127.0.0.1",
                "http": "0.0.0.0",
                "rpc": "127.0.0.1"
            }
        if self.servers:
            # TODO: check if we need to eliminate ourselves from the list?
            d["start_join"] = self.servers
        return d

    def _server_config(self):
        d = self._common_config()
        d['bootstrap_expect'] = self.bootstrap_expect
        d['server'] = True
        return d

    def _client_config(self):
        d = self._server_config()
        d['server'] = False
        return d

    def _upstart_conf(self):
        s = ''
        s += 'description "Consul %s process"\n\n' % self.mode
        s += 'start on (local-filesystems and net-device-up IFACE=eth0)\n'
        s += 'stop on runlevel [!12345]\n\n'
        s += 'respawn\n\n'
        s += 'setuid consul\nsetgid consul\n\n'
        s += 'exec /opt/consul/0.4.1/consul agent -config-dir /etc/consul.d\n'
        return s



#!/usr/bin/env python

import os
import pwd

from nose.tools import *

import carlcm

c = carlcm.Context()

def test_packages():
    try: c.shell('apt-get remove -y dtrx 2>&1 >>/dev/null', quiet=True)
    except: pass
    eq_(c.package('dtrx'), True)
    eq_(c.package('dtrx'), False)

def test_users():
    try: c.shell('userdel carltest', quiet=True)
    except: pass
    try: c.shell('groupdel carltest', quiet=True)
    except: pass
    try: c.shell('rm -rf /home/carltest', quiet=True)
    except: pass
    eq_(c.user('carltest'), True)
    eq_(c.user('carltest'), False)
    eq_(os.stat('/home/carltest').st_uid, pwd.getpwnam('carltest').pw_uid)

def test_groups():
    try: c.shell('groupdel carltestg', quiet=True)
    except: pass
    try: c.shell('userdel carltest', quiet=True)
    except: pass
    try: c.shell('groupdel carltest', quiet=True)
    except: pass
    try: c.shell('rm -rf /home/carltest', quiet=True)
    except: pass
    eq_(c.group('carltestg'), True)
    eq_(c.group('carltestg'), False)
    eq_(c.user('carltest'), True)
    eq_(c.user('carltest', groups=['carltestg']), True)
    eq_(c.user('carltest', groups=['carltestg']), False)

def test_z():
    # s = open(os.getenv('HOME') + '/.ssh/id_rsa.pub', 'rb').read().strip()
    s = 'ssh-rsa AAAAB3NzaC1yc2EAAAABIwAABAEAn8c1rMAb3yWrmtq+gtvFRh65qNliiLbSbPDTWmolIrM8slSMaIUXRRcJ6hHDoBmhGBHypQLtbVeYxFCEylhORNMHBYp5D1HhT03Tx/gLThlBXDTsxi78OFzzhAdsDYXviKtfziN/UKWNoY0WACln5xPCx7mHkgULzBQYzErqtgThVP1SWFBX0BzkJ2X1Wp6FFSX8dhmJSDg0VlioxvgSG15rrJ43EO6qrgAeSqzeAmBq5Kz3bPCloKRoxbuvNt5oWcC4Lnd5b5pOJIrGKF2IHNPZ7dvkjA3KvA91hq6I+/knrpNJQTHLnzRWaC/uT4cJ0nR3XALKZKsAXwRV4n/k6qZCWB+V1LWn9h6idZ1x/Xx0EL3Bo5HlqsDTreSWLFzRL7DyInmX1aVYlswNDgXWWofZ53W9WXbD6j9rRDX5Kt3+pw3vjORzcoEltDuUUgzDy+LbXj6A2z0iBgE73b49LuKomIC9qIC+8Ll6xU7Isy4dvAIagHq1BWNjD0PckaSWDIocP39RjtZ7R2aFXCyF/WxrZaINZXY/5HoB7dSmj0QTGAsEGEhuwNvwfe++Iu13BSf6uB/TjI5b3B7DSuM9H1Yyspp6+rzYicUYR8JxiZ9HRUm66/9hdp84c134Wg3GWEPyBBDYGGFgjm4gn7fdM0vWDvcVoMjvQ5rqx02yFe0oThN473Nm+ARMWjLailz7Vc7tKIBQgrx3/lnfxkjMnmyOhu0ONDtTkur4PHdThEmAN8apxAyrKF4Ejt65undyFscp3wr6f4xoREgYM9FV334EVhRbuktSbHxpHi0wXPe/S53GvytxcpZLgC1xNUDcB87BG7BAxcX4Bwv+CVlxuERQza8FRG9VC6QtVWPLASQG9RUdqKN0FWtKvhAW+M0dXXAHFa7rY7h8+uPeneZ+994btjslZX8PJO6ZueQvdBVVfMmJBB7BDeIHvTN6emAGJtInT9wJyxvSigSlKbxlPNR0g69QgZBkWGg4TBDrf2pVYWP0Jsi3WqJlRnE3L34+OW06+utADuXvB4p+cVmiCbx9ogV4I9EWA6y2TNPoG4pY83zrYKu9qgsmy5IjJe43rL58qzs9BHdVmD9cN5MfcpdAHdJgp6WxOi/K86BYsgFXQCxCZvW6AhelzhElp/W30Y0gFp5PRwaRgb4wQiJZILAxb2lJ1KXlIPNx/AbVgOJN7YvpEzNyBpKfTmaudwCwfVINLkvpaEkR8w0XnLrAFovWYv66pMn7AUIhiQWp+MhRvDzLGJmSiHOK3Pl22fzWHqbwrWKEpDPMYpYfYtay/5ktd9dshd6lyf9XVfktnrr3gotAIqv8vyi/xQstFOsGPWIVMU6cYLdWlEe0LRmFcw== carl@carlz'
    c.user('carl', authorized_keys=s, shell='/bin/bash')

'''
def test_download():
    c.shell('rm -f /opt/consul/0.4.1/consul.zip', quiet=True)
    eq_(c.download('/opt/consul/0.4.1/consul.zip', 'https://dl.bintray.com/mitchellh/consul/0.4.1_linux_amd64.zip', sha256sum='2cf6e59edf348c3094c721eb77436e8c789afa2c35e6e3123a804edfeb1744ac'), True)
    eq_(c.download('/opt/consul/0.4.1/consul.zip', 'https://dl.bintray.com/mitchellh/consul/0.4.1_linux_amd64.zip', sha256sum='2cf6e59edf348c3094c721eb77436e8c789afa2c35e6e3123a804edfeb1744ac'), False)
'''




import json
import os
import subprocess
import time

class Counselor(object):

    def __init__(self, region=None, ec2=None, iam=None, in_ec2=False):
        import boto
        import boto.ec2
        import boto.iam
        import consul

        if in_ec2:
            import boto.utils
            self.meta = boto.utils.get_instance_metadata()
        if region is None:
            if in_ec2:
                self.region = self.meta['hostname'].split('.')[1]
            else:
                self.region = 'us-west-2'
        else:
            self.region = region

        self.policies = {
            'counselor': {
                "Statement":[{
                    "Effect":"Allow",
                    "Action":["s3:*", "ec2:*"],
                "Resource":["*"]}]
            }
        }
        self.ec2 = ec2 or boto.ec2.connect_to_region(self.region)
        self.iam = iam or boto.connect_iam()
        assert self.ec2
        assert self.iam

    def role_and_profile(self, name):
        self.iam_role(name)
        self.instance_profile(name, name)

    def iam_role(self, name, policy_name='managedpolicy'):
        import boto
        policy = json.dumps(self.policies[name], sort_keys=True, indent=4, separators=(',', ': '))
        try: role = self.iam.get_role(name)
        except boto.exception.BotoServerError: role = None
        if role is None:
            self.iam.create_role(name)
            role = self.iam.get_role(name)
        # ...someday, maybe support for multiple policies via iam.list_role_policies()
        self.iam.put_role_policy(name, policy_name, policy)

    def instance_profile(self, name, role_name=None):
        import boto
        try: profile = self.iam.get_instance_profile(name)
        except boto.exception.BotoServerError: profile = None
        if profile is None:
            self.iam.create_instance_profile(name)
            profile = self.iam.get_instance_profile(name)
        roles = profile['get_instance_profile_response']['get_instance_profile_result']['instance_profile']['roles']
        should_add = False
        should_remove = False
        if role_name is None:
            if len(roles) != 0:
                should_remove = True
        if role_name is not None:
            if len(roles) == 0:
                should_add = True
            elif roles['member']['role_name'] != role_name:
                should_remove = True
                should_add = True
        if should_remove:
            self.iam.remove_role_from_instance_profile(name, roles['member']['role_name'])
        if should_add:
            self.iam.add_role_to_instance_profile(name, role_name)

        if role_name is not None and len(roles) != 0:
            should_add = True

    def security_group(self, name, description):
        import boto
        try: group = self.ec2.get_all_security_groups(groupnames=[name])[0]
        except boto.exception.EC2ResponseError: group = None
        if group is None:
            self.ec2.create_security_group(name, description)
            group = self.ec2.get_all_security_groups(groupnames=[name])[0]
        return group

    def get_instances(self):
        return reduce(lambda a, b: a + b, [a.instances for a in self.ec2.get_all_reservations()], [])

    def ensure_council(self):
        count = 3
        g1 = self.security_group('counselor', 'Security Group for Council').id
        g2 = self.security_group('counseled', 'Security Group for Instances Following to Council').id
        self.role_and_profile('counselor')
        insts = self.get_instances()
        counselors = [i for i in insts if 'counselor' in [g.name for g in i.groups]]
        counselors = [i for i in counselors if i.state in ['running', 'pending']]
        is_bootstrap = (len(counselors) == 0)
        user_data = '''#!/bin/bash
sudo apt-get update
sudo apt-get install -y python-dev python-setuptools
sudo easy_install pip
sudo pip install boto
sudo pip install python-consul
sudo pip install carlcm
sudo /usr/local/bin/carlcm-counselor %i''' % (count)
        if is_bootstrap:
            user_data += ' bootstrap'
        user_data += '\n'
        num_to_launch = count - len(counselors)
        d = {
            'image_id': 'ami-870a2fb7',
            'key_name': 'carl-ssh-2015-02-14',
            'user_data': user_data,
            'security_group_ids': [g1, g2],
            'subnet_id': 'subnet-7af8590d',
            'instance_type': 't2.micro',
            'instance_profile_name': 'counselor',
            'min_count': num_to_launch,
            'max_count': num_to_launch,
        }
        if num_to_launch > 0:
            self.ec2.run_instances(**d)
        return

    def ensure_local(self, count=1, is_bootstrap=False, context=None):
        from .context import Context
        from .modules import ConsulModule

        is_bootstrap = (is_bootstrap is True or is_bootstrap == 'bootstrap')

        instances = self.get_instances()

        counselors = [i for i in instances if 'counselor' in [g.name for g in i.groups]]
        counselors = [i for i in counselors if i.state in ['running', 'pending']]
        counselors = [i for i in counselors if i.private_ip_address != self.meta['local-ipv4']]
        counselor_ips = sorted([i.private_ip_address for i in counselors if i.private_ip_address])

        if is_bootstrap and all([self.meta['local-ipv4'] < ip for ip in counselor_ips]):
            counselor_ips = None

        c = Context()
        if context is not None:
            c = context

        c.add_modules(ConsulModule(mode='server', webui=True, bootstrap_expect=count, servers=counselor_ips))

        c.run_modules()

        if is_bootstrap:
            for i in xrange(100):
                time.sleep(5)
                try: subprocess.check_output(['service', 'consul', 'start'])
                except: pass



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

version = '0.1.12'

config = {
    'description': 'CarlCM',
    'author': 'Carl J Mackey',
    'url': 'https://github.com/cjmackey/carlcm',
    'author_email': 'carljmackey@gmail.com',
    'version': version,
    'install_requires': ['jinja2'],
    'packages': find_packages(),
    'scripts': [
        'bin/carlcm-bootstrap-council',
        'bin/carlcm-cluster-role',
        'bin/carlcm-counselor',
        'bin/carlcm-run-role',
    ],
    'name': 'carlcm',
    'classifiers': [
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: System :: Installation/Setup',
    ],
}

setup(**config)




from .base import BaseModule

class ExampleModule(BaseModule):

    def packages(self):
        return ['htop', 'dtrx', 'ack-grep']

    def main(self, context):
        context.user('carl')
        context.file('/home/carl/example.txt', src_data='hello carl\n',
                     owner='carl', group='carl')
